import asyncio

import pytest
from pytest_mock import MockerFixture

from aio_microservice import Service, ServiceSettings, amqp, http
from aio_microservice.amqp import AmqpExtension, AmqpExtensionSettings, TestAmqpBroker
from aio_microservice.http import TestHttpClient


async def test_amqp_asyncapi_ui() -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self, message: str) -> None:
            pass

    service = TestService()
    async with TestAmqpBroker(service=service), TestHttpClient(service=service) as http_client:
        response = await http_client.get("/schema/asyncapi")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert "test-subscriber-queue:_:HandleTest" in response.text


async def test_amqp_asyncapi_schema_download_json() -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self, message: str) -> None:
            pass

    service = TestService()
    async with TestAmqpBroker(service=service), TestHttpClient(service=service) as http_client:
        response = await http_client.get("/schema/asyncapi.json")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert "test-subscriber-queue:_:HandleTest" in response.json()["channels"]


async def test_amqp_asyncapi_schema_download_yaml() -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self, message: str) -> None:
            pass

    service = TestService()
    async with TestAmqpBroker(service=service), TestHttpClient(service=service) as http_client:
        response = await http_client.get("/schema/asyncapi.yaml")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert "test-subscriber-queue:_:HandleTest" in response.text


def test_amqp_asyncapi_schema_export_json(
    event_loop: asyncio.AbstractEventLoop,
    capsys: pytest.CaptureFixture[str],
    mocker: MockerFixture,
) -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self, message: str) -> None:
            pass

    mocker.patch.dict("os.environ", {"NO_COLOR": "1", "TERM": "dumb"})
    mocker.patch("sys.argv", ["test-service", "schema", "asyncapi", "--format", "json"])

    with pytest.raises(SystemExit):
        TestService.cli()

    captured = capsys.readouterr()
    assert "test-subscriber-queue:_:HandleTest" in captured.out


def test_amqp_asyncapi_schema_export_yaml(
    event_loop: asyncio.AbstractEventLoop,
    capsys: pytest.CaptureFixture[str],
    mocker: MockerFixture,
) -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self, message: str) -> None:
            pass

    mocker.patch.dict("os.environ", {"NO_COLOR": "1", "TERM": "dumb"})
    mocker.patch("sys.argv", ["test-service", "schema", "asyncapi", "--format", "yaml"])

    with pytest.raises(SystemExit):
        TestService.cli()

    captured = capsys.readouterr()
    assert "test-subscriber-queue:_:HandleTest" in captured.out


async def test_amqp_asyncapi_service_description() -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        __description__ = "TEST SERVICE DESCRIPTION"

    service = TestService()
    async with TestAmqpBroker(service=service), TestHttpClient(service=service) as http_client:
        response = await http_client.get("/schema/asyncapi.json")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert response.json()["info"]["description"] == TestService.__description__


async def test_amqp_asyncapi_service_version() -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        __version__ = "1.2.3"

    service = TestService()
    async with TestAmqpBroker(service=service), TestHttpClient(service=service) as http_client:
        response = await http_client.get("/schema/asyncapi.json")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert response.json()["info"]["version"] == TestService.__version__



from collections.abc import Iterator

import pytest
from _pytest.logging import LogCaptureFixture
from loguru import logger


# adopted from: https://loguru.readthedocs.io/en/stable/resources/migration.html#replacing-caplog-fixture-from-pytest-library
@pytest.fixture()
def caplog(caplog: LogCaptureFixture) -> Iterator[LogCaptureFixture]:
    handler_id = logger.add(
        caplog.handler,
        format="{message}",
        level=0,
        filter=lambda record: record["level"].no >= caplog.handler.level,
        enqueue=True,  # Set to 'True' if your test is spawning child processes.
    )
    yield caplog
    logger.remove(handler_id)



from faststream import FastStream
from faststream.asyncapi.generate import get_app_schema
from faststream.asyncapi.site import get_asyncapi_html
from litestar import Controller, MediaType, get


def make_asyncapi_controller(
    app: FastStream,
    path: str = "/schema",
) -> type[Controller]:
    routes_path = path

    class _AsyncAPIController(Controller):
        path: str = routes_path

        @get(
            path="/asyncapi",
            media_type=MediaType.HTML,
            include_in_schema=False,
            sync_to_thread=False,
        )
        def root(
            self,
            sidebar: bool = True,
            info: bool = True,
            servers: bool = True,
            operations: bool = True,
            messages: bool = True,
            schemas: bool = True,
            errors: bool = True,
            expand_message_examples: bool = True,
        ) -> str:
            schema = get_app_schema(app)
            return get_asyncapi_html(
                schema=schema,
                sidebar=sidebar,
                info=info,
                servers=servers,
                operations=operations,
                messages=messages,
                schemas=schemas,
                errors=errors,
                expand_message_examples=expand_message_examples,
                title=schema.info.title,
            )

        @get(
            path="/asyncapi.json",
            media_type=MediaType.JSON,
            include_in_schema=False,
            sync_to_thread=False,
        )
        def retrieve_schema_json(self) -> str:
            schema = get_app_schema(app)
            return schema.to_json()

        @get(
            path=["/asyncapi.yaml", "asyncapi.yml"],
            media_type="application/yaml",
            include_in_schema=False,
            sync_to_thread=False,
        )
        def retrieve_schema_yaml(self) -> str:
            schema = get_app_schema(app)
            return schema.to_yaml()

    return _AsyncAPIController



import pytest

from aio_microservice import (
    Service,
    ServiceSettings,
    startup_message,
)
from aio_microservice.http import TestHttpClient


async def test_startup_message(capsys: pytest.CaptureFixture[str]) -> None:
    class TestService(Service[ServiceSettings]):
        @startup_message
        async def test_startup_message(self) -> str:
            return "TEST STARTUP MESSAGE"

    service = TestService()
    async with TestHttpClient(service=service):
        captured = capsys.readouterr()
        assert "TEST STARTUP MESSAGE" in captured.out


async def test_startup_message_stacking(capsys: pytest.CaptureFixture[str]) -> None:
    class TestService(Service[ServiceSettings]):
        @startup_message
        async def test_startup_message_1(self) -> str:
            return "TEST STARTUP MESSAGE 1"

        @startup_message
        async def test_startup_message_2(self) -> str:
            return "TEST STARTUP MESSAGE 2"

    service = TestService()
    async with TestHttpClient(service=service):
        captured = capsys.readouterr()
        assert "TEST STARTUP MESSAGE 1" in captured.out
        assert "TEST STARTUP MESSAGE 2" in captured.out



from __future__ import annotations

from typing import TYPE_CHECKING

from aio_microservice import (
    Service,
    ServiceSettings,
    lifespan_hook,
    shutdown_hook,
    startup_hook,
)
from aio_microservice.http import TestHttpClient

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from pytest_mock import MockerFixture


async def test_startup_hook(mocker: MockerFixture) -> None:
    class TestService(Service[ServiceSettings]):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.startup_hook_stub = mocker.stub()

        @startup_hook
        async def test_startup_hook(self) -> None:
            self.startup_hook_stub()

    service = TestService()
    async with TestHttpClient(service=service):
        service.startup_hook_stub.assert_called_once()

    service.startup_hook_stub.assert_called_once()


async def test_startup_hook_stacking(mocker: MockerFixture) -> None:
    class TestService(Service[ServiceSettings]):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.startup_hook_1_stub = mocker.stub()
            self.startup_hook_2_stub = mocker.stub()

        @startup_hook
        async def test_startup_hook_1(self) -> None:
            self.startup_hook_1_stub()

        @startup_hook
        async def test_startup_hook_2(self) -> None:
            self.startup_hook_2_stub()

    service = TestService()
    async with TestHttpClient(service=service):
        service.startup_hook_1_stub.assert_called_once()
        service.startup_hook_2_stub.assert_called_once()

    service.startup_hook_1_stub.assert_called_once()
    service.startup_hook_2_stub.assert_called_once()


async def test_shutdown_hook(mocker: MockerFixture) -> None:
    class TestService(Service[ServiceSettings]):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.shutdown_hook_stub = mocker.stub()

        @shutdown_hook
        async def test_shutdown_hook(self) -> None:
            self.shutdown_hook_stub()

    service = TestService()
    async with TestHttpClient(service=service):
        service.shutdown_hook_stub.assert_not_called()

    service.shutdown_hook_stub.assert_called_once()


async def test_shutdown_hook_stacking(mocker: MockerFixture) -> None:
    class TestService(Service[ServiceSettings]):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.shutdown_hook_1_stub = mocker.stub()
            self.shutdown_hook_2_stub = mocker.stub()

        @shutdown_hook
        async def test_shutdown_hook_1(self) -> None:
            self.shutdown_hook_1_stub()

        @shutdown_hook
        async def test_shutdown_hook_2(self) -> None:
            self.shutdown_hook_2_stub()

    service = TestService()
    async with TestHttpClient(service=service):
        service.shutdown_hook_1_stub.assert_not_called()
        service.shutdown_hook_2_stub.assert_not_called()

    service.shutdown_hook_1_stub.assert_called_once()
    service.shutdown_hook_2_stub.assert_called_once()


async def test_lifespan_hook(mocker: MockerFixture) -> None:
    class TestService(Service[ServiceSettings]):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.startup_hook_stub = mocker.stub()
            self.shutdown_hook_stub = mocker.stub()

        @lifespan_hook
        async def test_lifespan_hook(self) -> AsyncGenerator[None, None]:
            self.startup_hook_stub()
            yield
            self.shutdown_hook_stub()

    service = TestService()
    async with TestHttpClient(service=service):
        service.startup_hook_stub.assert_called_once()
        service.shutdown_hook_stub.assert_not_called()

    service.startup_hook_stub.assert_called_once()
    service.shutdown_hook_stub.assert_called_once()


async def test_lifespan_hook_stacking(mocker: MockerFixture) -> None:
    class TestService(Service[ServiceSettings]):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.startup_hook_1_stub = mocker.stub()
            self.startup_hook_2_stub = mocker.stub()
            self.shutdown_hook_1_stub = mocker.stub()
            self.shutdown_hook_2_stub = mocker.stub()

        @lifespan_hook
        async def test_lifespan_hook_1(self) -> AsyncGenerator[None, None]:
            self.startup_hook_1_stub()
            yield
            self.shutdown_hook_1_stub()

        @lifespan_hook
        async def test_lifespan_hook_2(self) -> AsyncGenerator[None, None]:
            self.startup_hook_2_stub()
            yield
            self.shutdown_hook_2_stub()

    service = TestService()
    async with TestHttpClient(service=service):
        service.startup_hook_1_stub.assert_called_once()
        service.startup_hook_2_stub.assert_called_once()
        service.shutdown_hook_1_stub.assert_not_called()
        service.shutdown_hook_2_stub.assert_not_called()

    service.startup_hook_1_stub.assert_called_once()
    service.startup_hook_2_stub.assert_called_once()
    service.shutdown_hook_1_stub.assert_called_once()
    service.shutdown_hook_2_stub.assert_called_once()



from __future__ import annotations

import inspect
import logging

from loguru import logger as loguru_logger


class InterceptHandler(logging.Handler):
    """Default handler from examples in loguru documentation.

    See https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging.
    """

    def emit(self, record: logging.LogRecord) -> None:
        """Outputs the record through loguru.

        Args:
            record: The logging record.
        """
        # Get corresponding Loguru level if it exists.
        level: str | int
        try:
            level = loguru_logger.level(record.levelname).name
        except ValueError:  # pragma: no cover
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        loguru_logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def _get_intercepting_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.handlers = [InterceptHandler()]
    logger.propagate = False
    return logger


def _setup_warnings_logging() -> None:
    logging.captureWarnings(True)
    _get_intercepting_logger("py.warnings")


def setup_logging(level: int = logging.INFO) -> None:
    """Setup logging through loguru.

    This tries its best to force all logs through loguru.
    To achieve this an InterceptionHandler is registered.

    Further warnings will be forced into loguru.

    Args:
        level: The loglevel to apply to the root logger.
    """
    logging.basicConfig(handlers=[InterceptHandler()], level=level)

    # reconfigure existing loggers
    for name in logging.root.manager.loggerDict:
        _get_intercepting_logger(name)

    _setup_warnings_logging()



import pathlib

import pytest
from pytest_mock import MockerFixture

from aio_microservice import Service, ServiceSettings, http
from aio_microservice.http import TestHttpClient


async def test_http_openapi_swagger_ui() -> None:
    class TestService(Service[ServiceSettings]):
        @http.get(path="/test")
        async def get_test(self) -> str:
            return "TEST"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/schema/openapi")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert "/test" in response.text


async def test_http_openapi_schema_download_json() -> None:
    class TestService(Service[ServiceSettings]):
        @http.get(path="/test")
        async def get_test(self) -> str:
            return "TEST"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/schema/openapi.json")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert "/test" in response.json()["paths"]


async def test_http_openapi_schema_download_yaml() -> None:
    class TestService(Service[ServiceSettings]):
        @http.get(path="/test")
        async def get_test(self) -> str:
            return "TEST"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/schema/openapi.yaml")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert "/test" in response.text


def test_http_openapi_schema_export_json(
    capsys: pytest.CaptureFixture[str],
    mocker: MockerFixture,
) -> None:
    class TestService(Service[ServiceSettings]):
        @http.get(path="/test")
        async def get_test(self) -> str:
            return "TEST"

    mocker.patch.dict("os.environ", {"NO_COLOR": "1", "TERM": "dumb"})
    mocker.patch("sys.argv", ["test-service", "schema", "openapi", "--format", "json"])

    with pytest.raises(SystemExit):
        TestService.cli()

    captured = capsys.readouterr()
    assert "/test" in captured.out


def test_http_openapi_schema_export_with_output_json(
    tmp_path: pathlib.Path,
    mocker: MockerFixture,
) -> None:
    class TestService(Service[ServiceSettings]):
        @http.get(path="/test")
        async def get_test(self) -> str:
            return "TEST"

    output_file = tmp_path / "openapi.json"

    mocker.patch(
        "sys.argv",
        [
            "test-service",
            "schema",
            "openapi",
            "--format",
            "json",
            "--output",
            output_file,
        ],
    )

    with pytest.raises(SystemExit):
        TestService.cli()

    written_schema = output_file.read_text()
    assert "/test" in written_schema


def test_http_openapi_schema_export_yaml(
    capsys: pytest.CaptureFixture[str],
    mocker: MockerFixture,
) -> None:
    class TestService(Service[ServiceSettings]):
        @http.get(path="/test")
        async def get_test(self) -> str:
            return "TEST"

    mocker.patch.dict("os.environ", {"NO_COLOR": "1", "TERM": "dumb"})
    mocker.patch("sys.argv", ["test-service", "schema", "openapi", "--format", "yaml"])

    with pytest.raises(SystemExit):
        TestService.cli()

    captured = capsys.readouterr()
    assert "/test" in captured.out


async def test_http_openapi_service_description() -> None:
    class TestService(Service[ServiceSettings]):
        __description__ = "TEST SERVICE DESCRIPTION"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/schema/openapi.json")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert response.json()["info"]["description"] == TestService.__description__


async def test_http_openapi_service_version() -> None:
    class TestService(Service[ServiceSettings]):
        __version__ = "1.2.3"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/schema/openapi.json")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert response.json()["info"]["version"] == TestService.__version__



from typing import Any

from litestar import Request
from litestar.events import listener
from pytest_mock import MockerFixture

from aio_microservice import Service, ServiceSettings, http
from aio_microservice.http import TestHttpClient


async def test_http_event_listener(mocker: MockerFixture) -> None:
    listener_stub = mocker.stub()

    class TestService(Service[ServiceSettings]):
        @http.get(path="/test")
        async def get_test(self, request: Request[Any, Any, Any]) -> str:
            request.app.emit("get_test_called", self)
            return "TEST"

        @listener("get_test_called")
        async def on_get_test_called(self) -> None:
            listener_stub()

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/test")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert response.text == "TEST"

    listener_stub.assert_called_once()



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, ClassVar, TypeVar

import litestar

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Awaitable

    from pydantic import BaseModel


class CommonABC(ABC):
    __version__: ClassVar[str] = "1.0.0"
    __description__: ClassVar[str] = ""
    __http_controllers__: ClassVar[list[type[litestar.Controller]]] = []

    def __init__(self) -> None:
        self._startup_hooks: list[startup_hook] = []
        self._shutdown_hooks: list[shutdown_hook] = []
        self._lifespan_hooks: list[lifespan_hook] = []
        self._readiness_probes: list[readiness_probe] = []
        self._liveness_probes: list[liveness_probe] = []
        self._startup_messages: list[startup_message] = []
        self._schema_exports: list[schema_export] = []
        self._litestar_http_route_handlers: list[litestar.handlers.HTTPRouteHandler] = []
        self._litestar_http_controllers: list[litestar.types.ControllerRouterHandler] = []
        self._litestar_listeners: list[litestar.events.EventListener] = []

        # register controller classes
        self._litestar_http_controllers.extend(self.__http_controllers__)

        self._collect_decorated_functions()

    def _collect_decorated_functions(self) -> None:  # noqa: C901
        for _, attribute in inspect.getmembers(self.__class__):
            if isinstance(attribute, startup_hook):
                self._startup_hooks.append(attribute)
            elif isinstance(attribute, shutdown_hook):
                self._shutdown_hooks.append(attribute)
            elif isinstance(attribute, lifespan_hook):
                self._lifespan_hooks.append(attribute)
            elif isinstance(attribute, readiness_probe):
                self._readiness_probes.append(attribute)
            elif isinstance(attribute, liveness_probe):
                self._liveness_probes.append(attribute)
            elif isinstance(attribute, startup_message):
                self._startup_messages.append(attribute)
            elif isinstance(attribute, schema_export):
                self._schema_exports.append(attribute)
            elif isinstance(attribute, litestar.handlers.HTTPRouteHandler):
                self._litestar_http_route_handlers.append(attribute)
            elif isinstance(attribute, litestar.events.EventListener):
                self._litestar_listeners.append(attribute)


class ServiceABC(CommonABC):
    def __init__(self, settings: BaseModel) -> None:
        CommonABC.__init__(self)
        self._init_extensions(settings)

    def _init_extensions(self, settings: BaseModel) -> None:
        for extension_cls in ExtensionABC._extension_classes:
            if not isinstance(self, extension_cls):
                continue
            # pass settings if the extension-constructor expects them
            init_kwargs = {}
            init_signature = inspect.signature(extension_cls.__init__)
            if "settings" in init_signature.parameters:
                init_kwargs["settings"] = settings
            extension_cls.__init__(self, **init_kwargs)

    @abstractmethod
    async def run(self) -> None: ...  # pragma: no cover


class ExtensionABC(CommonABC):
    _extension_classes: ClassVar[set[type[ExtensionABC]]] = set()

    def __init_subclass__(cls) -> None:
        if ServiceABC in cls.__mro__:  # pragma: no cover
            return
        ExtensionABC._extension_classes.add(cls)

    def register_http_controller(
        self,
        controller: litestar.types.ControllerRouterHandler,
    ) -> None:
        self._litestar_http_controllers.append(controller)


CommonABCT = TypeVar("CommonABCT", bound=CommonABC)


class startup_hook:  # noqa: N801
    def __init__(
        self,
        fn: Callable[[CommonABCT], Awaitable[None]],
    ) -> None:
        self.fn = fn

    async def __call__(self, service: CommonABCT) -> None:
        await self.fn(service)


class shutdown_hook:  # noqa: N801
    def __init__(
        self,
        fn: Callable[[CommonABCT], Awaitable[None]],
    ) -> None:
        self.fn = fn

    async def __call__(self, service: CommonABCT) -> None:
        await self.fn(service)


class lifespan_hook:  # noqa: N801
    def __init__(
        self,
        fn: Callable[[CommonABCT], AsyncGenerator[None, None]],
    ) -> None:
        self.fn = fn


class readiness_probe:  # noqa: N801
    def __init__(
        self,
        fn: Callable[[CommonABCT], Awaitable[bool]],
    ) -> None:
        self.fn = fn

    async def __call__(self, service: CommonABCT) -> bool:
        return await self.fn(service)


class liveness_probe:  # noqa: N801
    def __init__(
        self,
        fn: Callable[[CommonABCT], Awaitable[bool]],
    ) -> None:
        self.fn = fn

    async def __call__(self, service: CommonABCT) -> bool:
        return await self.fn(service)


class startup_message:  # noqa: N801
    def __init__(
        self,
        fn: Callable[[CommonABCT], Awaitable[str]],
    ) -> None:
        self.fn = fn

    async def __call__(self, service: CommonABCT) -> str:
        return await self.fn(service)


class schema_export:  # noqa: N801
    MARKER = "_core_schema_export"

    def __init__(
        self,
        schema_type: str,
        schema_format: str,
    ) -> None:
        self.schema_type = schema_type
        self.schema_format = schema_format

    def __call__(
        self,
        fn: Callable[[CommonABCT], Awaitable[str]],
    ) -> schema_export:
        self.fn = fn
        return self



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Generic

from strawberry.litestar import BaseContext, make_graphql_controller
from strawberry.printer import print_schema

from aio_microservice import http
from aio_microservice.core.abc import (
    CommonABC,
    CommonABCT,
    ExtensionABC,
    schema_export,
    startup_message,
)

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterable

    from graphql import ExecutionResult as GraphQLExecutionResult
    from strawberry import Schema
    from strawberry.types import ExecutionResult
    from strawberry.types.graphql import OperationType


class GraphqlContext(Generic[CommonABCT], BaseContext):
    service: CommonABCT


class _SchemaWrapper:
    def __init__(self, service: GraphqlExtension) -> None:
        self._service = service
        self._schema = service.__graphql_schema__

    async def execute(
        self,
        query: str | None,
        variable_values: dict[str, Any] | None = None,
        root_value: Any | None = None,  # noqa: ANN401
        operation_name: str | None = None,
        allowed_operation_types: Iterable[OperationType] | None = None,
    ) -> ExecutionResult:
        context_value: GraphqlContext[CommonABC] = await self._service.graphql._context_getter()
        return await self._schema.execute(
            query=query,
            variable_values=variable_values,
            context_value=context_value,
            root_value=root_value,
            operation_name=operation_name,
            allowed_operation_types=allowed_operation_types,
        )

    async def subscribe(
        self,
        query: str,
        variable_values: dict[str, Any] | None = None,
        root_value: Any | None = None,  # noqa: ANN401
        operation_name: str | None = None,
    ) -> AsyncIterator[GraphQLExecutionResult] | GraphQLExecutionResult:
        context_value: GraphqlContext[CommonABC] = await self._service.graphql._context_getter()
        return await self._schema.subscribe(
            query=query,
            variable_values=variable_values,
            context_value=context_value,
            root_value=root_value,
            operation_name=operation_name,
        )


class GraphqlImpl:
    def __init__(self, service: GraphqlExtension) -> None:
        self._service = service
        self._graphql_controller = make_graphql_controller(
            schema=service.__graphql_schema__,
            path="/graphql",
            context_getter=self._context_getter,
        )

    async def _context_getter(self) -> GraphqlContext[CommonABCT]:
        return GraphqlContext(service=self._service)  # type: ignore

    @property
    def schema(self) -> _SchemaWrapper:
        return _SchemaWrapper(service=self._service)


class GraphqlExtension(ExtensionABC):
    __graphql_schema__: ClassVar[Schema]

    def __init__(self) -> None:
        self.graphql = GraphqlImpl(service=self)
        self.register_http_controller(self.graphql._graphql_controller)

    @startup_message
    async def _graphql_startup_message(self) -> str:
        return """
            # Graphql

            The following http-endpoints are available:
            - {scheme}://{host}:{port}/graphql
            - {scheme}://{host}:{port}/schema/graphql.graphql
        """

    @http.get(path="schema/graphql.graphql", include_in_schema=False)
    async def _graphql_get_schema(self) -> str:
        return print_schema(self.__graphql_schema__)

    @schema_export(schema_type="graphql", schema_format="graphql")
    async def _graphql_export_schema(self) -> str:
        return print_schema(self.__graphql_schema__)



from __future__ import annotations

from pydantic import BaseModel

from aio_microservice import Service, ServiceSettings, amqp, http
from aio_microservice.amqp import AmqpExtension, AmqpExtensionSettings, TestAmqpBroker
from aio_microservice.http import TestHttpClient


async def test_amqp_test_broker_subscriber_decorator() -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        def __init__(self, settings: TestSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.handler_got_called = False

        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self, message: str) -> None:
            self.handler_got_called = True

    service = TestService()

    async with TestAmqpBroker(service) as amqp_broker:
        await amqp_broker.publish(queue="test-subscriber-queue", message="TEST")
        assert service.handler_got_called is True


async def test_amqp_test_broker_publisher_decorator() -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        @amqp.subscriber(queue="test-subscriber-queue")
        @amqp.publisher(queue="test-publisher-queue")
        async def handle_test(self, message: str) -> str:
            return "TEST-RESPONSE"

    service = TestService()

    async with TestAmqpBroker(service) as amqp_broker:
        await amqp_broker.publish(queue="test-subscriber-queue", message="TEST-REQUEST")

        published_messages = amqp_broker.get_published_messages(queue="test-publisher-queue")
        assert len(published_messages) == 1
        assert published_messages[0] == "TEST-RESPONSE"


async def test_amqp_test_broker_publisher_decorator_pydantic() -> None:
    class Response(BaseModel):
        message: str
        number: int

    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        @amqp.subscriber(queue="test-subscriber-queue")
        @amqp.publisher(queue="test-publisher-queue")
        async def handle_test(self, message: str) -> Response:
            return Response(message="TEST-RESPONSE", number=42)

    service = TestService()

    async with TestAmqpBroker(service) as amqp_broker:
        await amqp_broker.publish(queue="test-subscriber-queue", message="TEST-REQUEST")

        published_messages = amqp_broker.get_published_messages(queue="test-publisher-queue")
        assert len(published_messages) == 1
        assert published_messages[0] == Response(message="TEST-RESPONSE", number=42).model_dump()


async def test_amqp_test_broker_publisher_instance() -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        def __init__(self, settings: TestSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.publisher = self.amqp.broker.publisher(
                queue="test-publisher-queue",
            )

        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self, message: str) -> None:
            await self.publisher.publish("TEST-RESPONSE")

    service = TestService()

    async with TestAmqpBroker(service) as amqp_broker:
        await amqp_broker.publish(queue="test-subscriber-queue", message="TEST-REQUEST")

        published_messages = amqp_broker.get_published_messages(queue="test-publisher-queue")
        assert len(published_messages) == 1
        assert published_messages[0] == "TEST-RESPONSE"


async def test_amqp_test_broker_publisher_instance_pydantic() -> None:
    class Response(BaseModel):
        message: str
        number: int

    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        def __init__(self, settings: TestSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.publisher = self.amqp.broker.publisher(
                queue="test-publisher-queue",
            )

        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self, message: str) -> None:
            response = Response(message="TEST-RESPONSE", number=42)
            await self.publisher.publish(response)

    service = TestService()

    async with TestAmqpBroker(service) as amqp_broker:
        await amqp_broker.publish(queue="test-subscriber-queue", message="TEST-REQUEST")

        published_messages = amqp_broker.get_published_messages(queue="test-publisher-queue")
        assert len(published_messages) == 1
        assert published_messages[0] == Response(message="TEST-RESPONSE", number=42).model_dump()


async def test_amqp_test_broker_reset_published_messages() -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        @amqp.subscriber(queue="test-subscriber-queue")
        @amqp.publisher(queue="test-publisher-queue")
        async def handle_test(self, message: str) -> str:
            return "TEST-RESPONSE"

    service = TestService()

    async with TestAmqpBroker(service) as amqp_broker:
        await amqp_broker.publish(queue="test-subscriber-queue", message="TEST-REQUEST")

        published_messages = amqp_broker.get_published_messages(queue="test-publisher-queue")
        assert len(published_messages) == 1

        amqp_broker.reset_published_messages()

        published_messages = amqp_broker.get_published_messages(queue="test-publisher-queue")
        assert len(published_messages) == 0

        await amqp_broker.publish(queue="test-subscriber-queue", message="TEST-REQUEST")

        published_messages = amqp_broker.get_published_messages(queue="test-publisher-queue")
        assert len(published_messages) == 1


async def test_amqp_test_broker_publisher_inside_http_handler() -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        def __init__(self, settings: TestSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.publisher = self.amqp.broker.publisher(
                queue="test-publisher-queue",
            )

        @http.get(path="/test")
        async def handle_test(self, message: str) -> str:
            response = "TEST-RESPONSE"
            await self.publisher.publish(response)
            return response

    service = TestService()
    async with TestAmqpBroker(service) as amqp_broker, TestHttpClient(service) as http_client:
        response = await http_client.get("/test?message=TEST-REQUEST")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert response.text == "TEST-RESPONSE"

        published_messages = amqp_broker.get_published_messages(queue="test-publisher-queue")

        assert len(published_messages) == 1
        assert published_messages[0] == "TEST-RESPONSE"



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from aio_microservice import Service, ServiceSettings, scheduler
from aio_microservice.http import TestHttpClient
from aio_microservice.scheduler import SchedulerExtension

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


async def test_scheduler_interval(mocker: MockerFixture) -> None:
    class TestService(Service[ServiceSettings], SchedulerExtension):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.do_every_second_stub = mocker.stub()

        @scheduler.interval(seconds=1)
        async def do_every_second(self) -> None:
            self.do_every_second_stub()

    service = TestService()
    async with TestHttpClient(service=service):
        await asyncio.sleep(2.5)

    assert 2 <= len(service.do_every_second_stub.call_args_list) <= 3


async def test_scheduler_cron(mocker: MockerFixture) -> None:
    class TestService(Service[ServiceSettings], SchedulerExtension):
        def __init__(self, settings: ServiceSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.do_every_second_stub = mocker.stub()

        @scheduler.cron(second="*/1")
        async def do_every_second(self) -> None:
            self.do_every_second_stub()

    service = TestService()
    async with TestHttpClient(service=service):
        await asyncio.sleep(2.5)

    assert 2 <= len(service.do_every_second_stub.call_args_list) <= 3


async def test_scheduler_crontab(mocker: MockerFixture) -> None:
    class TestService(Service[ServiceSettings], SchedulerExtension):
        @scheduler.crontab(expression="* * * * *")
        async def do_every_minute(self) -> None:
            pass

    service = TestService()

    jobs = service.scheduler.scheduler.get_jobs()
    assert len(jobs) == 1



from __future__ import annotations

import inspect
from datetime import timezone
from typing import TYPE_CHECKING, Callable, TypeVar

from apscheduler.schedulers.asyncio import AsyncIOScheduler  # type: ignore
from apscheduler.triggers.cron import CronTrigger  # type: ignore
from apscheduler.triggers.interval import IntervalTrigger  # type: ignore
from loguru import logger
from typing_extensions import Concatenate, ParamSpec

from aio_microservice.core.abc import ExtensionABC, shutdown_hook, startup_hook

if TYPE_CHECKING:
    from collections.abc import Awaitable


class SchedulerExtensionImpl:
    def __init__(self, service: SchedulerExtension) -> None:
        self._service = service
        self._scheduler = AsyncIOScheduler(timezone=timezone.utc)
        self._register_schedules()

    def _register_schedules(self) -> None:
        schedule_methods = inspect.getmembers(
            object=self._service,
            predicate=lambda obj: hasattr(obj, SchedulerDecorator.MARKER),
        )
        for schedule_name, _ in schedule_methods:
            schedule = getattr(self._service, schedule_name)
            schedule_settings = getattr(schedule, SchedulerDecorator.MARKER)
            for schedule_setting in schedule_settings:
                if isinstance(schedule_setting, interval):
                    self.add_interval(
                        fn=schedule,
                        weeks=schedule_setting.weeks,
                        days=schedule_setting.days,
                        hours=schedule_setting.hours,
                        minutes=schedule_setting.minutes,
                        seconds=schedule_setting.seconds,
                    )
                elif isinstance(schedule_setting, cron):
                    self.add_cron(
                        fn=schedule,
                        year=schedule_setting.year,
                        month=schedule_setting.month,
                        day=schedule_setting.day,
                        week=schedule_setting.week,
                        day_of_week=schedule_setting.day_of_week,
                        hour=schedule_setting.hour,
                        minute=schedule_setting.minute,
                        second=schedule_setting.second,
                    )
                elif isinstance(schedule_setting, crontab):  # pragma: no branch
                    self.add_crontab(
                        fn=schedule,
                        expression=schedule_setting.expression,
                    )

    @property
    def scheduler(self) -> AsyncIOScheduler:
        return self._scheduler

    def add_interval(
        self,
        fn: Callable[[], Awaitable[None]],
        weeks: int = 0,
        days: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: int = 0,
    ) -> None:
        trigger = IntervalTrigger(
            weeks=weeks,
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
        )
        self._scheduler.add_job(func=fn, trigger=trigger)

    def add_cron(
        self,
        fn: Callable[[], Awaitable[None]],
        year: int | str | None = None,
        month: int | str | None = None,
        day: int | str | None = None,
        week: int | str | None = None,
        day_of_week: int | str | None = None,
        hour: int | str | None = None,
        minute: int | str | None = None,
        second: int | str | None = None,
    ) -> None:
        trigger = CronTrigger(
            year=year,
            month=month,
            day=day,
            week=week,
            day_of_week=day_of_week,
            hour=hour,
            minute=minute,
            second=second,
        )
        self._scheduler.add_job(func=fn, trigger=trigger)

    def add_crontab(
        self,
        fn: Callable[[], Awaitable[None]],
        expression: str,
    ) -> None:
        trigger = CronTrigger.from_crontab(expr=expression)
        self._scheduler.add_job(func=fn, trigger=trigger)


class SchedulerExtension(ExtensionABC):
    def __init__(self) -> None:
        self.scheduler = SchedulerExtensionImpl(service=self)

    @startup_hook
    async def _scheduler_startup_hook(self) -> None:
        logger.info("Starting scheduler")
        self.scheduler._scheduler.start()

    @shutdown_hook
    async def _scheduler_shutdown_hook(self) -> None:
        logger.info("Stopping scheduler")
        self.scheduler._scheduler.shutdown()


SchedulerExtensionT = TypeVar("SchedulerExtensionT", bound=SchedulerExtension)
P = ParamSpec("P")
R = TypeVar("R")


class SchedulerDecorator:
    MARKER = "_scheduler_decorator"


class interval(SchedulerDecorator):  # noqa: N801
    def __init__(
        self,
        weeks: int = 0,
        days: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: int = 0,
    ) -> None:
        self.weeks = weeks
        self.days = days
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds

    def __call__(
        self,
        fn: Callable[Concatenate[SchedulerExtensionT, P], R],
    ) -> Callable[Concatenate[SchedulerExtensionT, P], R]:
        schedules = getattr(fn, self.MARKER, [])
        schedules.append(self)
        setattr(fn, self.MARKER, schedules)
        return fn


class cron(SchedulerDecorator):  # noqa: N801
    def __init__(
        self,
        year: int | str | None = None,
        month: int | str | None = None,
        day: int | str | None = None,
        week: int | str | None = None,
        day_of_week: int | str | None = None,
        hour: int | str | None = None,
        minute: int | str | None = None,
        second: int | str | None = None,
    ) -> None:
        self.year = year
        self.month = month
        self.day = day
        self.week = week
        self.day_of_week = day_of_week
        self.hour = hour
        self.minute = minute
        self.second = second

    def __call__(
        self,
        fn: Callable[Concatenate[SchedulerExtensionT, P], R],
    ) -> Callable[Concatenate[SchedulerExtensionT, P], R]:
        schedules = getattr(fn, self.MARKER, [])
        schedules.append(self)
        setattr(fn, self.MARKER, schedules)
        return fn


class crontab(SchedulerDecorator):  # noqa: N801
    def __init__(self, expression: str) -> None:
        self.expression = expression

    def __call__(
        self,
        fn: Callable[Concatenate[SchedulerExtensionT, P], R],
    ) -> Callable[Concatenate[SchedulerExtensionT, P], R]:
        schedules = getattr(fn, self.MARKER, [])
        schedules.append(self)
        setattr(fn, self.MARKER, schedules)
        return fn



from faststream import BaseMiddleware, context
from faststream.rabbit import RabbitBroker, RabbitExchange, RabbitQueue

from aio_microservice.amqp.extension import (
    AmqpExtension,
    AmqpExtensionSettings,
    AmqpSettings,
    publisher,
    subscriber,
)
from aio_microservice.amqp.testing import TestAmqpBroker

__all__ = [
    "AmqpExtension",
    "AmqpExtensionSettings",
    "AmqpSettings",
    "BaseMiddleware",
    "RabbitBroker",
    "RabbitExchange",
    "RabbitQueue",
    "TestAmqpBroker",
    "context",
    "publisher",
    "subscriber",
]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
import strawberry

from aio_microservice import Service, ServiceSettings, http
from aio_microservice.graphql import GraphqlExtension
from aio_microservice.http import TestHttpClient

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


async def test_graphql_schema_download() -> None:
    @strawberry.type
    class Query:
        @strawberry.field
        async def test(self) -> str:
            return "TEST RESPONSE"

    class TestService(Service[ServiceSettings], GraphqlExtension):
        __graphql_schema__ = strawberry.Schema(query=Query)

    service = TestService()
    async with TestHttpClient(service=service) as http_client:
        response = await http_client.get("/schema/graphql.graphql")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert "test: String!" in response.text


def test_graphql_schema_export(
    capsys: pytest.CaptureFixture[str],
    mocker: MockerFixture,
) -> None:
    @strawberry.type
    class Query:
        @strawberry.field
        async def test(self) -> str:
            return "TEST RESPONSE"

    class TestService(Service[ServiceSettings], GraphqlExtension):
        __graphql_schema__ = strawberry.Schema(query=Query)

    mocker.patch.dict("os.environ", {"NO_COLOR": "1", "TERM": "dumb"})
    mocker.patch("sys.argv", ["test-service", "schema", "graphql"])

    with pytest.raises(SystemExit):
        TestService.cli()

    captured = capsys.readouterr()
    assert captured.out == "type Query {\n  test: String!\n}\n"



from typing import Any

from litestar import MediaType, Request, get
from litestar.openapi import OpenAPIController as _OpenAPIController
from litestar.response.base import ASGIResponse


class OpenAPIController(_OpenAPIController):
    path = "/schema"

    @get(path="/openapi", include_in_schema=False, sync_to_thread=False)
    def swagger_ui(self, request: Request[Any, Any, Any]) -> ASGIResponse:
        return ASGIResponse(
            body=self.render_swagger_ui(request),
            media_type=MediaType.HTML,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import boto3
import botocore
from loguru import logger
from pydantic import BaseModel, Field, SecretStr

from aio_microservice.core.abc import ExtensionABC, readiness_probe, startup_hook

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client


class S3Settings(BaseModel):
    endpoint_url: str = Field(
        description="The URL to connect to.",
    )
    access_key_id: str = Field(
        description="The AWS access key ID used for authentication.",
    )
    secret_access_key: SecretStr = Field(
        description="The AWS secret access key used for authentication.",
    )


class S3ExtensionImpl:
    def __init__(self, settings: S3Settings) -> None:
        self._settings = settings
        self._boto3_session = boto3.Session(
            aws_access_key_id=self._settings.access_key_id,
            aws_secret_access_key=self._settings.secret_access_key.get_secret_value(),
        )
        self._boto3_s3_client = self.session.client(
            service_name="s3",
            endpoint_url=self._settings.endpoint_url,
        )

    @property
    def session(self) -> boto3.session.Session:
        return self._boto3_session

    @property
    def client(self) -> S3Client:
        return self._boto3_s3_client

    def verify_connection(self) -> bool:
        try:
            self._boto3_s3_client.list_buckets()
        except botocore.exceptions.EndpointConnectionError:
            pass
        else:
            return True
        return False


class S3ExtensionSettings(BaseModel):
    # NOTE the default values are required for schema-export
    s3: S3Settings = S3Settings(
        endpoint_url="http://dummy",
        access_key_id="dummy",
        secret_access_key="dummy",  # noqa: S106
    )


class S3Extension(ExtensionABC):
    def __init__(self, settings: S3ExtensionSettings) -> None:
        self.s3 = S3ExtensionImpl(settings=settings.s3)

    @startup_hook
    async def _s3_startup_hook(self) -> None:
        if not self.s3.verify_connection():
            logger.error("Failed to verify connection")

    @readiness_probe
    async def _s3_readiness_probe(self) -> bool:
        return self.s3.verify_connection()
